"""Visualize mesh files. Use the `--help` argument for information on how to
use the script."""

import argparse
from functools import reduce
from os import path
from tempfile import gettempdir

import numpy as np
from pyrr import Matrix44

from .. import Mesh, Lines, Scene, TexturedMesh
from ..behaviours import SceneInit
from ..behaviours.keyboard import SnapshotOnKey
from ..behaviours.misc import LightToCamera
from ..utils import save_frame
try:
    from ..window import show
    no_gui = False
except ImportError:
    no_gui = True


# The tab20 colormap from matplotlib
tab20 = [
    (0.12156862745098039, 0.4666666666666667,  0.7058823529411765  ),
    (0.6823529411764706,  0.7803921568627451,  0.9098039215686274  ),
    (1.0,                 0.4980392156862745,  0.054901960784313725),
    (1.0,                 0.7333333333333333,  0.47058823529411764 ),
    (0.17254901960784313, 0.6274509803921569,  0.17254901960784313 ),
    (0.596078431372549,   0.8745098039215686,  0.5411764705882353  ),
    (0.8392156862745098,  0.15294117647058825, 0.1568627450980392  ),
    (1.0,                 0.596078431372549,   0.5882352941176471  ),
    (0.5803921568627451,  0.403921568627451,   0.7411764705882353  ),
    (0.7725490196078432,  0.6901960784313725,  0.8352941176470589  ),
    (0.5490196078431373,  0.33725490196078434, 0.29411764705882354 ),
    (0.7686274509803922,  0.611764705882353,   0.5803921568627451  ),
    (0.8901960784313725,  0.4666666666666667,  0.7607843137254902  ),
    (0.9686274509803922,  0.7137254901960784,  0.8235294117647058  ),
    (0.4980392156862745,  0.4980392156862745,  0.4980392156862745  ),
    (0.7803921568627451,  0.7803921568627451,  0.7803921568627451  ),
    (0.7372549019607844,  0.7411764705882353,  0.13333333333333333 ),
    (0.8588235294117647,  0.8588235294117647,  0.5529411764705883  ),
    (0.09019607843137255, 0.7450980392156863,  0.8117647058823529  ),
    (0.6196078431372549,  0.8549019607843137,  0.8980392156862745  )
]


def list_of(f):
    def inner(x):
        try:
            return [f(xi) for xi in x.split(";")]
        except ValueError as e:
            raise ValueError("Error when parsing the list of elements") from e
    return inner

def or_type(f1, f2):
    def inner(x):
        try:
            return f1(x)
        except ValueError:
            return f2(x)
    return inner


def int_tuple(n):
    def inner(x):
        t = tuple(int(xi) for xi in x.split(","))
        if len(t) != n:
            raise ValueError("Expected a {}-tuple".format(n))
        return t
    return inner


def f_tuple(n):
    def inner(x):
        t = tuple(float(xi) for xi in x.split(","))
        if len(t) != n:
            raise ValueError("Expected a {}-tuple".format(n))
        return t
    return inner


def scene_init(args):
    def inner(scene):
        scene.camera_position = args.camera_position
        scene.camera_target = args.camera_target
        scene.up_vector = args.up
        scene.light = args.light or scene.camera_position
        if args.depth:
            H, W = args.orthographic_bounds
            scene.camera_matrix = Matrix44.orthogonal_projection(
                left=-W/2, right=W/2,
                bottom=H/2, top=-H/2,
                near=args.orthographic_near,
                far=args.orthographic_far
            )
        if args.show_axes:
            scene.add(Lines.axes())
    return inner


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Visualize meshes with simple_3dviz"
    )
    parser.add_argument(
        "file",
        default="The mesh to be viewed",
        nargs="+"
    )
    parser.add_argument(
        "--size",
        type=int_tuple(2),
        default=(512, 512),
        help="The size of the window"
    )
    parser.add_argument(
        "--background", "-b",
        type=or_type(f_tuple(4), f_tuple(3)),
        default=(0.7, 0.7, 0.7, 1),
        help="The rgba background color"
    )
    parser.add_argument(
        "--camera_position", "-c",
        type=f_tuple(3),
        default=(-2, -2, -2),
        help="The position of the camera"
    )
    parser.add_argument(
        "--camera_target", "-t",
        type=f_tuple(3),
        default=(0, 0, 0),
        help="The target of the camera"
    )
    parser.add_argument(
        "--up",
        type=f_tuple(3),
        default=(0, 0, 1),
        help="The up vector"
    )
    parser.add_argument(
        "--light",
        type=f_tuple(3),
        default=None
    )
    parser.add_argument(
        "--color",
        type=list_of(or_type(f_tuple(4), f_tuple(3))),
        default=[(0.3, 0.3, 0.3)],
        help="Choose a color for the mesh"
    )
    parser.add_argument(
        "--use_tab20",
        action="store_true",
        help="Use matplotlib's tab20 color map"
    )
    parser.add_argument(
        "--manual",
        action="store_false",
        dest="auto",
        help="Auto determine the camera position and target"
    )
    parser.add_argument(
        "--save_frame",
        default=path.join(gettempdir(), "frame_{:03d}.png"),
        help="The location to save the snapshot frame"
    )
    parser.add_argument(
        "--direct_render",
        help="If provided render to this file and exit"
    )
    parser.add_argument(
        "--with_texture",
        action="store_true",
        help="Load mesh with texture"
    )
    parser.add_argument(
        "--depth",
        action="store_true",
        help="Use an orthographic camera and render the depth of the scene"
    )
    parser.add_argument(
        "--orthographic_bounds",
        type=f_tuple(2),
        help=("When an orthographic projection is used, these define the "
              "width and height of the image plane in the world.")
    )
    parser.add_argument(
        "--orthographic_near",
        type=float,
        default=0.1,
        help="Set nearest visible point to the camera"
    )
    parser.add_argument(
        "--orthographic_far",
        type=float,
        default=10,
        help="Set farthest visible point to the camera"
    )
    parser.add_argument(
        "--show_axes",
        action="store_true",
        help="Draw the axes at (0, 0, 0)"
    )

    args = parser.parse_args(argv)

    MeshClass = TexturedMesh if args.with_texture else Mesh
    colors = args.color*len(args.file) if len(args.color) == 1 else args.color
    if args.use_tab20:
        colors = [tab20[i % 20] for i in range(len(args.file))]
    meshes = [
        MeshClass.from_file(f, color=c)
        for f, c in zip(args.file, colors)
    ]

    # If manual is not set then move the camera far enough so that it can see
    # the whole scene
    if args.auto:
        bbox_min = reduce(
            np.minimum,
            (m.bbox[0] for m in meshes),
            meshes[0].bbox[0]
        )
        bbox_max = reduce(
            np.maximum,
            (m.bbox[1] for m in meshes),
            meshes[0].bbox[1]
        )
        bbox = [bbox_min, bbox_max]
        center = (bbox[1]-bbox[0])/2 + bbox[0]
        args.camera_target = center
        args.camera_position = center + (bbox[1]-center)*2
        # In case the camera is too far, then scale the objects so that the
        # camera distance is not very large.
        # NOTE: This probably only works with a single model
        D = np.sqrt(np.sum((args.camera_position - args.camera_target)**2))
        if D > 100:
            s = 100. / D
            args.camera_target *= s
            args.camera_position *= s
            for m in meshes:
                m.scale(s)

    # Fill the orthographic_bounds if they are not given by computing the scene
    # bounding box and setting the camera plane to be a square with a side
    # equal to the diagonal of the bbox of the scene.
    if args.orthographic_bounds is None:
        bbox_min = reduce(
            np.minimum,
            (m.bbox[0] for m in meshes),
            meshes[0].bbox[0]
        )
        bbox_max = reduce(
            np.maximum,
            (m.bbox[1] for m in meshes),
            meshes[0].bbox[1]
        )
        diagonal = np.sqrt(((bbox_max - bbox_min)**2).sum())
        args.orthographic_bounds = (diagonal, diagonal)

    # Set the meshes to render in depth mode if args.depth is set
    if args.depth:
        for m in meshes:
            if hasattr(m, "mode"):
                m.mode = "depth"

    behaviours = [
        SnapshotOnKey(path=args.save_frame, keys={"<ctrl>", "S"}),
    ]
    if args.light is None:
        behaviours.append(LightToCamera())

    if args.direct_render:
        scene = Scene(size=args.size, background=args.background)
        for m in meshes:
            m.sort_triangles(args.camera_position)
            scene.add(m)
        scene_init(args)(scene)
        scene.render()
        save_frame(args.direct_render, scene.frame)
    else:
        show(
            meshes,
            size=args.size, background=args.background, title="Mesh Viewer",
            camera_position=args.camera_position,
            camera_target=args.camera_target,
            up_vector=args.up,
            light=args.light,
            behaviours=[SceneInit(scene_init(args))] + behaviours
        )



"""Load a mesh into three arrays vertices, faces and vertex_colors."""

import numpy as np
import os
from plyfile import PlyData

from ._utils import get_file, close_file


class MeshReader(object):
    """MeshReader defines the common interface for reading all supported mesh
    files."""
    def __init__(self, filename=None):
        self._vertices = None
        self._normals = None
        self._colors = None
        self._uv = None
        self._material_file = None

        if filename is not None:
            self.read(filename)

    def read(self, filename):
        raise NotImplementedError()

    @property
    def vertices(self):
        if self._vertices is None:
            raise NotImplementedError()
        return self._vertices

    @property
    def normals(self):
        if self._normals is None:
            raise NotImplementedError()
        return self._normals

    @property
    def colors(self):
        if self._colors is None:
            raise NotImplementedError()
        return self._colors

    @property
    def uv(self):
        if self._uv is None:
            raise NotImplementedError()
        return self._uv

    @property
    def material_file(self):
        if self._material_file is None:
            raise NotImplementedError()
        return self._material_file


class PlyMeshReader(MeshReader):
    """Read binary or ascii PLY files."""
    def __init__(self, filename=None, vertex="vertex", face="face",
                 vertex_indices=None):
        self._names = {
            "vertex": vertex,
            "face": face,
            "vertex_indices": vertex_indices
        }
        super(PlyMeshReader, self).__init__(filename)

    def _get_colors(self, p):
        colors = np.zeros((p.count, 4), dtype=np.float32)
        colors[:, 0] = p["red"]
        colors[:, 1] = p["green"]
        colors[:, 2] = p["blue"]
        colors[:, 3] = 255
        try:
            colors[:, 3] = p["alpha"]
        except ValueError:
            pass
        colors /= 255
        return colors

    def read(self, filename):
        # Make local copies of the names of the ply elements
        V = self._names["vertex"]
        F = self._names["face"]
        VI = self._names["vertex_indices"]

        # Read the file into a PlyData datastucture
        data = PlyData.read(filename)

        # Get the vertices
        if V in data:
            self._vertices = np.vstack([
                data[V]["x"],
                data[V]["y"],
                data[V]["z"]
            ]).T

            # If we have per vertex colors get them as well
            per_vertex_colors = None
            try:
                per_vertex_colors = self._get_colors(data[V])
            except ValueError:
                pass
        if F in data:
            if VI is None:
                VI = data[F].properties[0].name
            triangles = sum(
                len(indices)-2
                for indices in data[F][VI]
            )
            if triangles == len(data[F][VI]):
                faces = np.vstack(data[F][VI])
            else:
                raise NotImplementedError(("Dealing with non-triangulated "
                                           "faces is not supported"))
            self._vertices = self._vertices[faces].reshape(-1, 3)
            if per_vertex_colors is not None:
                self._colors = per_vertex_colors[faces].reshape(-1, 4)
            else:
                try:
                    colors = self._get_colors(data[F])
                    self._colors = np.repeat(colors, 3, axis=0)
                except ValueError:
                    pass


class ObjMeshReader(MeshReader):
    """Read OBJ mesh files."""
    def _triangulate_faces(self, faces):
        triangles = []
        for f in faces:
            if len(f) == 3:
                triangles.append(f)
            else:
                for i in range(2, len(f)):
                    triangles.append([f[0], f[i-1], f[i]])
        return triangles

    def read(self, filename):
        def extract_vertex(face):
            return int(face.split("/")[0])-1

        def extract_normal(face):
            return int(face.split("/")[2])-1

        def extract_uv(face):
            return int(face.split("/")[1])-1

        try:
            f = get_file(filename)

            lines = f.readlines()

            # Collect all the vertices, namely lines starting with 'v' followed
            # by 3 floats and arrange them according to faces
            vertices = np.array([
                list(map(float, l.strip().split()[1:4]))
                for l in lines if l.startswith("v ")
            ], dtype=np.float32)
            faces = np.array(self._triangulate_faces([
                list(map(extract_vertex, l.strip().split()[1:]))
                for l in lines if l.startswith("f ")
            ]))
            self._vertices = vertices[faces].reshape(-1, 3)

            # Collect all the vertex normals, namely lines starting with 'vn'
            # followed by 3 floats and arrange the according to faces
            try:
                normals = np.array([
                    list(map(float, l.strip().split()[1:]))
                    for l in lines if l.startswith("vn ")
                ])
                faces = np.array(self._triangulate_faces([
                    list(map(extract_normal, l.strip().split()[1:]))
                    for l in lines if l.startswith("f ")
                ]))
                self._normals = normals[faces].reshape(-1, 3)
            except IndexError:
                pass

            # Collect all the texture coordinates, namely u, v [,w] coordinates
            try:
                uv = np.array([
                    list(map(float, l.strip().split()[1:]))
                    for l in lines if l.startswith("vt ")
                ])
                faces = np.array(self._triangulate_faces([
                    list(map(extract_uv, l.strip().split()[1:]))
                    for l in lines if l.startswith("f ")
                ]))
                self._uv = uv[faces].reshape(-1, uv.shape[1])[:, :2]
            except IndexError:
                pass

            # Collect a material file
            try:
                material_file = [
                    l.strip().split()[1:][0]
                    for l in lines if l.startswith("mtllib")
                ][0]

                if isinstance(filename, str):
                    self._material_file = os.path.join(
                        os.path.dirname(filename),
                        material_file
                    )
                else:
                    if hasattr(filename, "name"):
                        self._material_file = os.path.join(
                            os.path.dirname(filename.name),
                            material_file
                        )
                    else:
                        pass
            except IndexError:
                pass
        finally:
            close_file(filename, f)


class OffMeshReader(MeshReader):
    """Read OFF mesh files."""
    def read(self, filename):
        try:
            f = get_file(filename)

            # Read lines and clean them from comments and empty lines
            lines = f.readlines()
            lines = [l.strip() for l in lines if l[0]!="#" and l.strip() != ""]

            # Ensure it is an OFF file
            if not lines[0].startswith("OFF"):
                raise ValueError("Invalid OFF file.")

            # Extract the number of vertices, faces and edges
            if len(lines[0].split()) > 1:
                n_vertices, n_faces, n_edges = [
                    int(x)
                    for x in lines[0].split()[1:]
                ]
                lines = lines[1:]
            else:
                n_vertices, n_faces, n_edges = [
                    int(x)
                    for x in lines[1].split()
                ]
                lines = lines[2:]

            # Collect the vertices and faces
            vertices = np.array([
                [float(x) for x in l.split()]
                for l in lines[:n_vertices]
            ], dtype=np.float32)
            faces = np.array([
                [float(x) for x in l.split()]
                for l in lines[n_vertices:n_vertices+n_faces]
            ], dtype=np.float32)

            if not np.all(faces[:, 0] == 3):
                raise NotImplementedError(("Dealing with non-triangulated "
                                           "faces is not supported"))

            # Set the triangles
            vertex_indices = faces[:, 1:4].astype(int).ravel()
            self._vertices = vertices[:, :3][vertex_indices]

            # Set the colors
            if vertices.shape[1] > 3:
                self._colors = vertices[:, 4:][vertex_indices]
            elif faces.shape[1] > 4:
                self._colors = np.repeat(faces[:, 4:], 3, axis=0)
        finally:
            close_file(filename, f)


class StlMeshReader(MeshReader):
    """Read STL mesh files."""
    def read(self, filename):
        # Decide if it is an ASCII STL or not
        ascii_stl = True
        try:
            f = get_file(filename, "rb")
            try:
                past_header = str(f.read(100), encoding="ascii")
            except UnicodeDecodeError:
                ascii_stl = False
            f.seek(0)

            if ascii_stl:
                vertices = []
                normals = []
                START_SOLID = 1
                START_FACE = 2
                START_VERTEX = 3
                END_VERTEX = 4
                END_FACE = 5
                END_SOLID = 6
                normal = None
                state = START_SOLID
                for line in f:
                    line = str(line, encoding="ascii")
                    fields = line.strip().split()
                    if state == START_SOLID:
                        if fields[0] != "solid":
                            raise ValueError("Non ASCII STL")
                        state = START_FACE
                    elif state == START_FACE:
                        if fields[0] == "endsolid":
                            state = START_SOLID
                            continue
                        else:
                            assert fields[0] == "facet" and fields[1] == "normal"
                            normal = [float(x) for x in fields[2:5]]
                            state = START_VERTEX
                    elif state == START_VERTEX:
                        if fields[0] == "outer" and fields[1] == "loop":
                            continue
                        elif fields[0] == "vertex":
                            v = [float(x) for x in fields[1:4]]
                            vertices.append(v)
                            normals.append(normal)
                        elif fields[0] == "endloop":
                            state = END_FACE
                    elif state == END_FACE:
                        if fields[0] == "outer" and fields[1] == "loop":
                            state = START_VERTEX
                        else:
                            assert fields[0] == "endfacet"
                            state = START_FACE
                self._vertices = np.array(vertices)
                self._normals = np.array(normals)
            else:
                header = f.read(80)
                assert len(header) == 80
                triangles = np.frombuffer(f.read(4), "<i4", 1)[0]
                dtype = np.dtype([
                    ("normal", "<f4", (3,)),
                    ("vertices", "<f4", (3, 3)),
                    ("attr", "<u2")
                ])
                mesh = np.frombuffer(f.read(), dtype, triangles)
                self._vertices = mesh["vertices"].reshape(-1, 3)
                self._normals = np.repeat(mesh["normal"], 3, 0).reshape(-1, 3)
        finally:
            close_file(filename, f)



def _get_file(filename, mode="r"):
    if isinstance(filename, str):
        return open(filename, mode)
    return filename


def _close_file(filename, f):
    """Close the file if filename is a string."""
    if hasattr(f, "close") and isinstance(filename, str):
        f.close()



"""Behaviours decouple the changes of a scene with the management of a scene
(creation, rendering, etc.).

The behaviours allow for a simple and unified way to animate, save frames to
disk, interact with a mouse and keyboard. This means that you can write all
your domain logic in terms of behaviours and then render with or without a GUI
by using `simple_3dviz.window.show` or `simple_3dviz.utils.render`
respectively.
"""

class Behaviour(object):
    """Behaviour defines the interface that is implemented by all behaviours
    and provides simple helper classes to encapsulate the current scene
    configuration.

    In particular the interface is a single method
    `simple_3dviz.behaviours.Behaviour.behave` that accepts a single argument
    of type `simple_3dviz.behaviours.Behaviour.Params`. For details see the
    corresponding comments.
    """
    class Mouse(object):
        """Hold information about the mouse events.

        Example behaviours that use the mouse can be found in
        `simple_3dviz.behaviours.mouse`.

        Attributes
        ----------
            location: (x, y), a tuple of floats or integers containing the
                      location of the mouse in the screen
            left_pressed: bool, indicating whether the left button of the mouse
                          is currently pressed (or has been pressed once)
            middle_pressed: bool, indicating whether the middle button of the
                            mouse is currently pressed (or has been pressed
                            once)
            wheel_rotation: float | int, indicate the number of position
                            changes of the scroll wheel of the mouse, the sign
                            indicates the direction of rotation
        """
        def __init__(self, location, left_pressed, middle_pressed, wheel_rotation):
            self.location = location
            self.left_pressed = left_pressed
            self.middle_pressed = middle_pressed
            self.wheel_rotation = wheel_rotation

    class Keyboard(object):
        """Holds information about the keyboard events.

        The members are two sets containing string representations for the keys
        that were pressed down and released respectively. Example behaviours
        that use the keyboard can be found in
        `simple_3dviz.behaviours.keyboard` and
        `simple_3dviz.scripts.mesh_viewer`.

        Attributes
        ----------
            keys_down: set[string], the keys that were pressed down
            keys_up: set[string], the keys that were released
        """
        def __init__(self, keys_down, keys_up):
            self.keys_down = set(keys_down)
            self.keys_up = set(keys_up)

    class Params(object):
        """The parameters provided by the window to the behaviours.

        This object facilitates interaction between behaviours the same way an
        event object facilitates interactions between DOM event handlers in
        languages like javascript.

        The properties `stop_propagation`, `done` and `refresh` are used to
        stop handling of behaviours for this tick, remove the current behaviour
        from the behaviours list and cause a repaint event respectively.

        Attributes
        ----------
            window: A simple_3dviz.window.base.BaseWindow implementation
                    pointing to the current window object
            scene: A simple_3dviz.scenes.Scene object pointing to the current
                   scene
            frame: A Callable that returns the current frame
            mouse: A simple_3dviz.behaviours.Behaviour.Mouse object providing
                   mouse info
            keyboard: A simple_3dviz.behaviours.Behaviour.Keyboard object
                      providing keyboard info
            last_call: bool, when true this means that the rendering has
                       finished and the behaviours are called to clean up
            stop_propagation: bool, when set no more behaviours will be run
            done: bool, when set remove this behaviour from the list
            refresh: bool, when set make sure to redraw the window
        """
        def __init__(self, window, scene, frame, mouse, keyboard, last_call):
            self.window = window
            self.scene = scene
            self.frame = frame
            self.mouse = mouse
            self.keyboard = keyboard
            self.last_call = last_call

            self._stop = False
            self._done = False
            self._refresh = False

        @property
        def stop_propagation(self):
            """Set to True to stop the processing of more behaviours in the
            current tick."""
            return self._stop

        @stop_propagation.setter
        def stop_propagation(self, s):
            self._stop = s

        @property
        def done(self):
            """Set to True to remove this behaviour from the behaviours
            list."""
            return self._done

        @done.setter
        def done(self, d):
            self._done = d

        @property
        def refresh(self):
            """Set to True to force a repaint event."""
            return self._refresh

        @refresh.setter
        def refresh(self, r):
            self._refresh = self._refresh or (r == True)

    def behave(self, params):
        """Implements the logic of the behaviour.

        Subclasses of behaviour must override this method to implement their
        custom logic.

        Arguments
        ---------
            params: Instance of simple_3dviz.behaviours.Behaviour.Params
        """
        raise NotImplementedError()


class SceneInit(Behaviour):
    """Initialize a scene.

    Run an init function once and update the render.

    Arguments
    ---------
        scene_init: callable, a function that takes a Scene as an argument to
                    initialize it, add objects, set the camera, etc.
    """
    def __init__(self, scene_init):
        self._init_func = scene_init

    def behave(self, params):
        self._init_func(params.scene)
        params.done = True
        params.stop_propagation = True
        params.refresh = True



"""Create a window with an attached scene that is rendered in real time."""

try:
    from .wx import Window
except ImportError:
    raise ImportError("No supported gui library was found. Install wxPython.")


from ..renderables import Renderable
from ..behaviours import Behaviour, SceneInit
from ..behaviours.mouse import MouseRotate, MouseZoom, MousePan
from ..behaviours.keyboard import CameraReport, SortTriangles


def simple_window(init, size=(512, 512)):
    """Return a window with the expected behaviours added.

    Arguments
    ---------
        init: callable that sets up the scene
        size: (w, h) the size of the window

    Returns
    -------
        Window instance
    """
    w = Window(size)
    w.add_behaviours([SceneInit(init), MouseRotate(), MouseZoom(),
                      MousePan(), CameraReport(), SortTriangles()])
    return w


def show(renderables, size=(512, 512), background=(1,)*4, title="Scene",
         camera_position=(-2, -2, -2), camera_target=(0, 0, 0),
         up_vector=(0, 0, 1), light=None, behaviours=[]):
    """Creates a simple window that displays the renderables.

    Arguments
    ---------
        renderables: list[Renderable] the renderables to be displayed in the
                     scene
        size: (w, h) the size of the window
        background: (r, g, b, a) the rgba tuple for the background
        title: str the title of the window
        camera_position: (x, y, z) the position of the camera
        camera_target: (x, y, z) the point that the camera looks at
        up_vector: (x, y, z) defines the floor and sky
        light: (x, y, z) defines the position of the light source
    """
    if not isinstance(renderables, (list, tuple)):
        renderables = [renderables]
    if not all(isinstance(r, Renderable) for r in renderables):
        raise ValueError(("show() expects one or more renderables as "
                          "parameters not {}").format(renderables))

    def init(scene):
        for r in renderables:
            scene.add(r)
        scene.background = background
        scene.camera_position = camera_position
        scene.camera_target = camera_target
        scene.up_vector = up_vector
        if light is not None:
            scene.light = light

    w = simple_window(init, size=size)
    w.add_behaviours(behaviours)
    w.show(title)




import numpy as np

from . import Behaviour


class LightToCamera(Behaviour):
    """Move the light to the same position as the camera so that the objects
    are always lit."""
    def __init__(self, offset=[0, 0, 0]):
        self._offset = offset

    def behave(self, params):
        newpos = params.scene.camera_position + self._offset
        if np.allclose(newpos, params.scene.light):
            return

        params.scene.light = newpos
        params.refresh = True


class AddObjectsSequentially(Behaviour):
    """Add a series of objects to a scene at constant time intervals.

    Arguments
    ---------
        objects: A list of Renderable objects
        interval: The interval between additions of objects in ticks
    """
    def __init__(self, objects, interval=30):
        self._objects = objects
        self._interval = interval

        self._ticks = interval
        self._index = len(objects)

    def behave(self, params):
        self._ticks += 1
        if self._ticks > self._interval:
            self._ticks = 0
            if self._index >= len(self._objects):
                for o in self._objects:
                    params.scene.remove(o)
                self._index = 0
            params.scene.add(self._objects[self._index])
            self._index += 1
            params.refresh = True


class CycleThroughObjects(Behaviour):
    """Add a set of objects to the scene removing the ones previously added.

    Arguments
    ---------
        objects: A list of lists of Renderable objects
        interval: The interval between additions and removals in ticks
    """
    def __init__(self, objects, interval=30):
        self._objects = objects
        self._interval = interval

        self._ticks = interval
        self._object = -1

    def behave(self, params):
        self._ticks += 1
        if self._ticks > self._interval:
            self._ticks = 0
            for o in self._objects[self._object]:
                params.scene.remove(o)
            self._object = (self._object + 1) % len(self._objects)
            for o in self._objects[self._object]:
                params.scene.add(o)
            params.refresh = True


class SortTriangles(Behaviour):
    """Sort the triangles of the scene wrt the camera position."""
    def __init__(self):
        self._prev_camera = None

    def behave(self, params):
        camera = params.scene.camera_position
        if (
            self._prev_camera is None or
            any([
                abs(c1-c2)>1e-3 for c1, c2 in zip(self._prev_camera, camera)
            ])
        ):
            self._prev_camera = camera
            for r in params.scene.renderables:
                if hasattr(r, "sort_triangles"):
                    r.sort_triangles(camera)
            params.refresh = True


class StartStopBehaviour(Behaviour):
    """Start and stop the decorated behaviour for the given frame counts."""
    def __init__(self, behaviour, start=0, stop=100):
        self._cnt = 0
        self._behaviour = behaviour
        self._start = start
        self._stop = stop

    def behave(self, params):
        self._cnt += 1
        if self._stop >= self._cnt > self._start:
            return self._behaviour.behave(params)



"""Visualize two-dimensional functions. Use the `--help` argument for
information on how to use the script."""

import argparse

import matplotlib.pyplot as plt
import numpy as np

from .. import Mesh, Lines
from ..window import show


def int_tuple(n):
    def inner(x):
        t = tuple(int(xi) for xi in x.split(","))
        if len(t) != n:
            raise ValueError("Expected a {}-tuple".format(n))
        return t
    return inner


def f_tuple(n):
    def inner(x):
        t = tuple(float(xi) for xi in x.split(","))
        if len(t) != n:
            raise ValueError("Expected a {}-tuple".format(n))
        return t
    return inner


def get_colormap(cmap, log_colors):
    cmap = plt.cm.get_cmap(cmap)
    def normalize(x):
        xmin = x.min()
        xmax = x.max()
        return (x-xmin)/(xmax-xmin)

    def colors(x):
        if log_colors:
            return cmap(normalize(np.log(np.maximum(x, 1e-7))))
        else:
            return cmap(normalize(x))

    return colors


def get_function(func, xlim, ylim, n, cmap="viridis", log_colors=False):
    x = np.linspace(xlim[0], xlim[1], n)
    y = np.linspace(ylim[0], ylim[1], n)
    X, Y = np.meshgrid(x, y)
    Z = eval(func, dict(x=X, y=Y, np=np))

    return Mesh.from_xyz(X, Y, Z, colormap=get_colormap(cmap, log_colors))


def get_axes():
    return Lines(
        [[-1, -1, -1],
         [ 1, -1, -1],
         [-1, -1, -1],
         [-1,  1, -1],
         [-1, -1, -1],
         [-1, -1,  1]],
        [[0.5, 0.5, 0.5, 1.],
         [0.5, 0.5, 0.5, 1.],
         [0.5, 0.5, 0.5, 1.],
         [0.5, 0.5, 0.5, 1.],
         [0.5, 0.5, 0.5, 1.],
         [0.5, 0.5, 0.5, 1.]],
        width=0.01
    )


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Visualize functions with simple_3dviz"
    )
    parser.add_argument(
        "function",
        default="The function to be viewed as a python string (use x, y and np)"
    )
    parser.add_argument(
        "--n_points",
        type=int,
        default=100,
        help="How many points per dimension"
    )
    parser.add_argument(
        "--xlim",
        type=f_tuple(2),
        default=(-1., 1.),
        help="The limits for the x axis"
    )
    parser.add_argument(
        "--ylim",
        type=f_tuple(2),
        default=(-1., 1.),
        help="The limits for the y axis"
    )
    parser.add_argument(
        "--colormap",
        default="jet",
        help="Set the matplotlib colormap"
    )
    parser.add_argument(
        "--log_colors",
        action="store_true",
        help="Use logspace for assigning the colors"
    )
    parser.add_argument(
        "--size",
        type=int_tuple(2),
        default=(512, 512),
        help="The size of the window"
    )
    parser.add_argument(
        "--background", "-b",
        type=f_tuple(4),
        default=(0.7, 0.7, 0.7, 1),
        help="The rgba background color"
    )
    parser.add_argument(
        "--camera_position", "-c",
        type=f_tuple(3),
        default=(3, 3, 3),
        help="The position of the camera"
    )
    parser.add_argument(
        "--camera_target", "-t",
        type=f_tuple(3),
        default=(0, 0, 0),
        help="The target of the camera"
    )
    parser.add_argument(
        "--up",
        type=f_tuple(3),
        default=(0, 0, 1),
        help="The up vector"
    )
    parser.add_argument(
        "--light",
        type=f_tuple(3),
        default=(-0.5, -0.8, -2)
    )
    parser.add_argument(
        "--no_axes",
        action="store_false",
        dest="axes",
        help="Do not show the axes"
    )

    args = parser.parse_args(argv)

    mesh = get_function(args.function, args.xlim, args.ylim,
                        args.n_points, args.colormap, args.log_colors)
    axes = get_axes()
    show(
        [mesh] + ([axes] if args.axes else []),
        size=args.size, background=args.background, title="Func Viewer",
        camera_position=args.camera_position, camera_target=args.camera_target,
        up_vector=args.up, light=args.light
    )



"""Provide a common interface for loading mesh files or other useful files."""

from os import path

from .mesh import ObjMeshReader, OffMeshReader, PlyMeshReader, StlMeshReader
from .material import MtlMaterialReader


def read_mesh_file(filename, ext=None):
    if isinstance(filename, str):
        _, ext = path.splitext(filename)
    try:
        return {
            ".ply": PlyMeshReader,
            ".obj": ObjMeshReader,
            ".off": OffMeshReader,
            ".stl": StlMeshReader
        }[ext.lower()](filename)
    except KeyError as e:
        raise ValueError("{} mesh file extension is not supported".format(ext))


def read_material_file(filename, ext=None):
    if isinstance(filename, str):
        _, ext = path.splitext(filename)
    try:
        return {
            ".mtl": MtlMaterialReader
        }[ext.lower()](filename)
    except KeyError as e:
        raise ValueError("{} material file extension is not supported".format(ext))



"""Simple visualization of meshes using opengl with moderngl."""

__maintainer__ = "Angelos Katharopoulos, Despoina Paschalidou"
__email__ = "katharas@gmail.com"
__url__ = "http://simple-3dviz.com"
__description__ = "simple-3dviz is a simple visualization library for 3D "
"data using Python and OpenGL."
__keywords__ = "graphics geometry 3D"
__license__ = "MIT"
__version__ = "0.5.0"

from .renderables import Mesh, TexturedMesh, Spherecloud, Lines
from .scenes import Scene
from .utils import render



from .behaviours import Behaviour
from .renderables import Renderable
from .scenes import Scene


try:
    import cv2
    import numpy as np

    def save_frame(path, frame):
        channels = frame.shape[-1]
        if channels == 1:
            cv2.imwrite(path, frame)

        # swap channels from rgb to bgr
        image_data = [frame[:, :, i] for i in range(channels)]
        image_data[0], image_data[2] = image_data[2], image_data[0]
        frame = np.stack(image_data, axis=-1)

        cv2.imwrite(path, frame)

    def read_image(path):
        img = cv2.imread(path, -1)
        channels = img.shape[-1]
        if channels == 1:
            return img

        # swap channels from bgr to rgb
        image_data = [img[:, :, i] for i in range(channels)]
        image_data[0], image_data[2] = image_data[2], image_data[0]
        return np.stack(image_data, axis=-1)

except ImportError:
    from PIL import Image
    import numpy as np

    def save_frame(path, frame):
        Image.fromarray(frame).save(path)

    def read_image(path):
        return np.asarray(Image.open(path))


def render(renderables, behaviours, n_frames, size=(512, 512),
           background=(1,)*4, camera_position=(-2, -2, -2),
           camera_target=(0, 0, 0), up_vector=(0, 0, 1), light=None,
           scene=None):
    """Render a list of primitives for a given number of frames calling the
    passed behaviours after every frame.

    Arguments
    ---------
        renderables: list[Renderable] the renderables to be displayed in the
                     scene
        behaviours: list[Behaviour] a list of behaviours to animate the scene
                    and save the results
        n_frames: int the number of frames to render
        size: (w, h) the size of the window
        background: (r, g, b, a) the rgba tuple for the background
        camera_position: (x, y, z) the position of the camera
        camera_target: (x, y, z) the point that the camera looks at
        up_vector: (x, y, z) defines the floor and sky
        light: (x, y, z) defines the position of the light source
        scene: An optional scene to reuse
    """
    # Create the scene or clear it if it is provided
    if scene is None:
        scene = Scene(size=size)
    else:
        scene.clear()

    # Set up the scene
    scene.background = background
    scene.camera_position = camera_position
    scene.camera_target = camera_target
    scene.up_vector = up_vector
    scene.light = light if light is not None else camera_position

    # Add the primitives
    if not isinstance(renderables, (list, tuple)):
        renderables = [renderables]
    if not all(isinstance(r, Renderable) for r in renderables):
        raise ValueError(("render() expects one or more renderables as "
                          "parameters not {}").format(renderables))
    for r in renderables:
        scene.add(r)

    # Render the frames and run the behaviours
    # TODO: The following code duplicates some logic that can be found also in
    #       the implementations of BaseWindow. Investigate whether this could
    #       be refactored out.
    for frame in range(n_frames):
        # render
        scene.render()

        # run the behaviours
        params = Behaviour.Params(
            None,                 # we have no window
            scene,                # the scene
            lambda: scene.frame,  # return the frame if needed
            None,                 # no mouse
            None,                 # no keyboard
            frame == n_frames-1   # is this the last frame?
        )
        remove = []
        for i, b in enumerate(behaviours):
            b.behave(params)
            if params.done:
                remove.append(i)
                params.done = False
            if params.stop_propagation:
                break
        for i in reversed(remove):
            behaviours.pop(i)


def normalize_colors(c, N):
    """Normalize `c` so that it represents N 4-dimensional color values."""
    c = np.asarray(c)
    if len(c.shape) == 1:
        if c.shape[0] == 3:
            c = np.array(c.tolist() + [1.0])
        c = c[np.newaxis].repeat(N, axis=0)
    elif len(c.shape) == 2:
        if len(c) != N:
            raise ValueError(("c cannot represent the requested number "
                              "of colors"))
        if c.shape[1] == 3:
            c = np.hstack([c, np.ones((N, 1))])

    if c.shape != (N, 4):
        raise ValueError(("c cannot represent the requested number of colors"))

    return c.astype(np.float32)



"""Load a material file into ambient, diffuse and specular colors and a set of
texture maps."""

import numpy as np
import os

from .utils import _get_file, _close_file
from ..utils import read_image


class MaterialReader(object):
    """MaterialReader defines the common interface for reading material files.
    """
    def __init__(self, filename=None):
        self._materials = {}
        self._current = None

        if filename is not None:
            self.read(filename)

    def read(self, filename):
        raise NotImplementedError()


    def set_material(self, name):
        self._current = name

    @property
    def material_names(self):
        return sorted(list(self._materials.keys()))

    @property
    def ambient(self):
        return self._materials[self._current]["ambient"]

    @property
    def diffuse(self):
        return self._materials[self._current]["diffuse"]

    @property
    def specular(self):
        return self._materials[self._current]["specular"]

    @property
    def Ns(self):
        return self._materials[self._current]["Ns"]

    @property
    def texture(self):
        return self._materials[self._current]["texture"]

    @property
    def bump_map(self):
        return self._materials[self._current]["bump"]

    @property
    def mode(self):
        return self._materials[self._current]["mode"]


class MtlMaterialReader(MaterialReader):
    """Read MTL material files."""
    def _read_texture(self, filename, texture_path):
        if isinstance(filename, str):
            return read_image(os.path.join(
                os.path.dirname(filename),
                texture_path
            ))
        else:
            if hasattr(filename, "name"):
                return read_image(os.path.join(
                    os.path.dirname(filename.name),
                    texture_path
                ))
            else:
                return None

    def read(self, filename):
        try:
            f = None
            f = _get_file(filename)
            lines = f.readlines()

            materials = {}
            material = None
            for l in lines:
                l = l.strip()

                # Parse the name of the new mtl and add the previous to the
                # material list
                if l.startswith("newmtl"):
                    if material is not None:
                        materials[material["name"]] = material
                    material = {
                        "name": l.split()[1],
                        "ambient": np.array([1., 1, 1]),
                        "diffuse": np.array([1., 1, 1]),
                        "specular": np.array([0., 0, 0]),
                        "mode": "specular",
                        "Ns": 10.0,
                        "bump": None,
                        "texture": None
                    }

                # Collect the material color information, namely lines starting
                # with "Ka" followed by 3 floats indicating the r,g,b
                # components for the ambient reflectivity, "Kd" followed by 3
                # floats indicating the r,g,b components for the diffuse
                # reflectivity and "Ks" followed by 3 floats indicating the
                # r,g,b components for the specular reflectivity.
                elif l.startswith("Ka"):
                    material["ambient"] = np.array(
                        list(map(float, l.split()[1:]))
                    )

                elif l.startswith("Kd"):
                    material["diffuse"] = np.array(
                        list(map(float, l.split()[1:]))
                    )

                elif l.startswith("Ks"):
                    material["specular"] = np.array(
                        list(map(float, l.split()[1:]))
                    )

                # Collect the specular exponent, namely a line starting with
                # "Ns" followed by a float.
                elif l.startswith("Ns"):
                    material["Ns"] = float(l.split()[1])

                # Read the diffuse texture image
                elif l.startswith("map_Kd"):
                    material["texture"] = self._read_texture(
                        filename,
                        l.split()[1]
                    )

                # Read the bump map
                elif l.startswith("bump"):
                    material["bump"] = self._read_texture(
                        filename,
                        l.split()[1]
                    )

                # Collect the illumination model used for the material, namely
                # a line starting with "illum" and followed by an integer,
                # specifying the id of the illumination model.
                elif l.startswith("illum"):
                    material["mode"] = {
                        "0" : "constant",
                        "1" : "diffuse",
                        "2" : "specular"
                    }[l.split()[1]]

            if material is not None:
                materials[material["name"]] = material
                self._current = material["name"]

            self._materials = materials

        finally:
            if f is not None:
                _close_file(filename, f)



class Renderable(object):
    """Represent something that can be rendered."""
    def init(self, ctx):
        """Initialize the renderable with the moderngl context."""
        raise NotImplementedError()

    def release(self):
        """Release all resources of this renderable in relation to open gl"""
        raise NotImplementedError()

    def update_uniforms(self, uniforms):
        """Update any uniforms that are provided from somewhere else.

        The object is free to update its own programs' uniform values as well.
        Also since the program can be shared, the uniforms might be updated
        from another object.

        Arguments
        ---------
            uniforms: list of tuples
        """
        pass

    def render(self):
        """Render whatever this represents."""
        raise NotImplementedError()





import numpy as np

from .base import Renderable


class Spherecloud(Renderable):
    def __init__(self, centers, colors=(0.3, 0.3, 0.3), sizes=(0.02)):
        self._centers = np.asarray(centers)
        self._colors = np.asarray(colors)
        self._sizes = np.asarray(sizes)

        N = len(self._centers)
        if len(self._colors.shape) == 1:
            if self._colors.size == 3:
                self._colors = np.array(self._colors.tolist() + [1])
            self._colors = self._colors[np.newaxis].repeat(N, axis=0)
        elif self._colors.shape[1] == 3:
            self._colors = np.hstack([self._colors, np.ones((N, 1))])

        if len(self._sizes.shape) == 0:
            self._sizes = np.ones(N)*self._sizes

        self._prog = None
        self._vbo = None
        self._vao = None

    @property
    def packed_parameters(self):
        # Define the triangular pyramid assuming radius 1 and center 0 and then
        # offset and scale according to centers and sizes
        r2 = np.sqrt(2)
        r3 = np.sqrt(3)
        pyramid = np.array([[ 0,   r3,       0],
                            [ 1,    0,       0],
                            [-1,    0,       0],
                            [ 0, 1/r3, 2*r2/r3]])
        center = pyramid.mean(axis=0)
        ab = pyramid[1]-pyramid[0]
        ac = pyramid[2]-pyramid[0]
        normal = np.cross(ab, ac)
        normal /= np.sqrt(np.dot(normal, normal))
        max_radius = np.abs(np.dot(normal, pyramid[0]-center))
        pyramid -= center
        pyramid /= max_radius
        pyramid_vertices = pyramid[[0, 1, 2,
                                    0, 1, 3,
                                    0, 2, 3,
                                    1, 2, 3]][np.newaxis]

        vertices = self._sizes[:, np.newaxis, np.newaxis] * pyramid_vertices
        vertices += self._centers[:, np.newaxis]
        vertices = vertices.reshape(-1, 3)
        centers = np.repeat(self._centers, 12, axis=0)
        colors = np.repeat(self._colors, 12, axis=0)
        radii = np.repeat(self._sizes[:, np.newaxis], 12, axis=0)

        return np.hstack([vertices, centers, colors, radii]).astype(np.float32)

    def init(self, ctx):
        self._prog = ctx.program(
            vertex_shader="""
                #version 330

                uniform mat4 mvp;
                in vec3 in_vertex;
                in vec3 in_center;
                in vec4 in_color;
                in float in_radius;
                out vec3 v_vertex;
                out vec3 v_center;
                out vec4 v_color;
                out float v_radius;

                void main() {
                    v_vertex = in_vertex;
                    v_center = in_center;
                    v_color = in_color;
                    v_radius = in_radius;
                    gl_Position = mvp * vec4(in_vertex, 1);
                }
            """,
            fragment_shader="""
                #version 330

                uniform mat4 vm;
                uniform vec3 light;
                in vec3 v_vertex;
                in vec3 v_center;
                in vec4 v_color;
                in float v_radius;
                out vec4 f_color;

                void main() {
                    vec3 camera_position = vm[3].xyz / vm[3].w;
                    vec3 center_ray = v_center - camera_position;
                    vec3 ray = normalize(v_vertex - camera_position);
                    float tc = dot(center_ray, ray);
                    if (tc < 0) {
                        discard;
                    }
                    float d = sqrt(dot(center_ray, center_ray) - tc*tc);
                    if (d > v_radius) {
                        discard;
                    }
                    float t1c = sqrt(v_radius*v_radius - d*d);
                    vec3 p = camera_position + ray * (tc-t1c);

                    float lum = dot(
                        normalize(p - v_center),
                        normalize(p - light)
                    );
                    lum = acos(lum) / 3.14159265;
                    lum = clamp(lum, 0.0, 1.0);

                    f_color = vec4(v_color.xyz * lum, v_color.w);
                }
            """
        )
        self._vbo = ctx.buffer(self.packed_parameters.tobytes())
        self._vao = ctx.simple_vertex_array(
            self._prog,
            self._vbo,
            "in_vertex", "in_center", "in_color", "in_radius"
        )

    def sort_triangles(self, point):
        """Sort the triangles wrt point from further to closest."""
        centers = self._centers
        colors = self._colors
        sizes = self._sizes

        d = ((np.asarray(point).reshape(1, 3) - centers)**2).sum(-1)
        alpha = (colors[:, ::4].mean(-1)<1).astype(np.float32) * 1000
        idxs = np.argsort(d+alpha)[::-1]

        self._centers = centers[idxs]
        self._colors = colors[idxs]
        self._sizes = sizes[idxs]
        self._vbo.write(self.packed_parameters.tobytes())

    def release(self):
        self._prog.release()
        self._vbo.release()
        self._vao.release()

    def render(self):
        self._vao.render()

    def update_uniforms(self, uniforms):
        for k, v in uniforms:
            if k in ["light", "mvp", "vm"]:
                self._prog[k].write(v.tobytes())

    @property
    def bbox(self):
        """The axis aligned bounding box of all the vertices as two
        3-dimensional arrays containing the minimum and maximum for each
        axis."""
        return [
            self._centers.min(axis=0),
            self._centers.max(axis=0)
        ]

    def scale(self, s):
        """Multiply all the vertices with a number s."""
        self._centers *= s
        if self._vbo is not None:
            self._vbo.write(self.packed_parameters.tobytes())

    def to_unit_cube(self):
        """Transform the mesh such that it fits in the 0 centered unit cube."""
        bbox = self.bbox
        dims = bbox[1] - bbox[0]
        self._centers -= dims/2 + bbox[0]
        self._centers /= dims.max()
        if self._vbo is not None:
            self._vbo.write(self.packed_parameters.tobytes())
